unit GRuler;

(*****************************************************************************
 * GRuler 1.0 by Greg Butler
 *
 * Horizontal and vertical rulers in one control.  I wrote this because none
 * of the rulers I downloaded looked the way I wanted or did what I needed.
 *
 * The rulers are sizeable and scale automatically pretty well.  They can
 * show the current cursor position on the ruler with minimal coding.  They
 * support English and Metric measure as well as custom units.
 *
 ****************************************************************************)

interface

uses
  Windows, SysUtils, Classes, Graphics, Controls, Forms ;

const
  DefRulerTextColor = clWindowText ;
  DefRulerColor = clYellow ;
  DefDivisions = 5 ;

type
  TRulerOrientation = (roHorizontal, roVertical) ;
  TMeasureUnit      = (muInches, muCentimeters, muCustom) ;
  TRulerEdge        = (reBottom, reRight, reTop, reLeft) ;

  TGRuler = class(TGraphicControl)
  private
    { Private declarations }
    FPixelsPerUnit  : Integer ;
    FOffset         : Integer ; // Offset in pixels
    FDivisions      : Integer ; // Number of unit divisions
    FOrientation    : TRulerOrientation ;
    FUnit           : TMeasureUnit ;
    FEdge           : TRulerEdge ;
    FTickColor      : TColor ;
    FShowPosition   : Boolean ;
    FAlignment      : TAlignment ;
    PositionShowing : Boolean ;
    LastPosition    : Integer ;
    procedure SetEdge(const Value: TRulerEdge);
    procedure SetMeasureUnit(const Value: TMeasureUnit);
    procedure SetOffset(const Value: Integer);
    procedure SetOrientation(const Value: TRulerOrientation);
    procedure SetPixelsPerUnit(const Value: Integer);
    procedure SetTickColor(const Value: TColor);
    function StorePixelsPerUnit: Boolean ;
    procedure SetDivisions(const Value: Integer);
    procedure SetShowPosition(const Value: Boolean);
    procedure SetAlignment(const Value: TAlignment);
  protected
    { Protected declarations }
    procedure Paint; override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override ;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure UpdatePosition(const X, Y : Integer) ;
  published
    { Published declarations }
    property Align ;
    property Anchors ;
    property Color default DefRulerColor ;
    property Font ;
    property ParentColor ;
    property ParentFont ;
    property ParentShowHint ;
    property PopupMenu ;
    property ShowHint ;
    property Visible ;
    property Alignment : TAlignment read FAlignment write SetAlignment ;
    property Divisions : Integer read FDivisions write SetDivisions
                                 default DefDivisions ;
    property Edge : TRulerEdge read FEdge write SetEdge ;
    property Offset : Integer read FOffset write SetOffset ;
    property Orientation : TRulerOrientation read FOrientation
                                             write SetOrientation ;
    property PixelsPerUnit : Integer read FPixelsPerUnit
                                     write SetPixelsPerUnit
                                     stored StorePixelsPerUnit ;
    property ShowPosition : Boolean read FShowPosition
                                    write SetShowPosition ;
    property TickColor : TColor read FTickColor
                                write SetTickColor default DefRulerTextColor ;
    property MeasureUnit : TMeasureUnit read FUnit write SetMeasureUnit ;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Moon', [TGRuler]);
end;

{ TGRuler }

constructor TGRuler.Create(AOwner: TComponent);
begin
  inherited Create(AOwner) ;
  Height := 41 ;
  Width := 600 ;
  Color := DefRulerColor ;
  FTickColor := DefRulerTextColor ;
  Font.Color := DefRulerTextColor ;
  FDivisions := DefDivisions ;
  MeasureUnit := FUnit ;
end;

procedure TGRuler.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y) ;
  if(FShowPosition) then
    UpdatePosition(X, Y) ;
end;

procedure TGRuler.Paint;
var
  i, X, Y, iEdge, iSize, iDir, iMax, iOffset, iPadding : Integer ;
  PAxis, PEdge : PInteger ;
  d : Double ;
  S : String ;
  TextSizeFunc : function(const Text : String) : Integer of object ;
begin
  iEdge := 0 ;
  iSize := 0 ;
  iDir := 0 ;
  iPadding := 0 ;
  iOffset := 0 ;
  if(FOrientation = roHorizontal) then
  begin
    PAxis := @X ;
    PEdge := @Y ;
    iMax := Width ;
    TextSizeFunc := Canvas.TextWidth ;
  end
  else
  begin
    PAxis := @Y ;
    PEdge := @X ;
    iMax := Height ;
    TextSizeFunc := Canvas.TextHeight ;
    iPadding := 5 ;
  end ; { if }
  case FAlignment of
    taLeftJustify : iOffset := FOffset ;
    taCenter : iOffset := (iMax DIV 2) + FOffset ;
    taRightJustify : iOffset := iMax + FOffset ;
  end ; { case }
  case Edge of
    reBottom :
      begin
        iEdge := Height ;
        iDir := -1 ;
        iSize := Height ;
      end ;
    reRight :
      begin
        iEdge := Width ;
        iDir := -1 ;
        iSize := Width ;
      end ;
    reTop :
      begin
        iEdge := 0 ;
        iDir := 1 ;
        iSize := Height ;
      end ;
    reLeft :
      begin
        iEdge := 0 ;
        iDir := 1 ;
        iSize := Width ;
      end ;
  end ;{ case }
  if(FUnit = muCentimeters) then
    d := Screen.PixelsPerInch / (2.54 * FDivisions)
  else
    d := FPixelsPerUnit / FDivisions ;
  Canvas.Font.Assign(Font) ;
  Canvas.Font.Size := 4 ;
  while(Canvas.TextHeight('8') < (iSize DIV 2)) do
    Canvas.Font.Size := Canvas.Font.Size + 1 ;
  Canvas.Pen.Color := Font.Color ;
  Canvas.Brush.Color := Color ;
  Canvas.Brush.Style := bsSolid ;
  Canvas.Rectangle(0, 0, Width, Height);
  Canvas.Brush.Style := bsClear ;
  Canvas.Pen.Color := FTickColor ;
  i := Trunc(iOffset / d) * -1 ;
  repeat
    PAxis^ := iOffset + Trunc(i * d) ;
    PEdge^ := iEdge ;
    Canvas.MoveTo(X, Y);
    if((i MOD FDivisions) = 0) then
    begin
      Inc(PEdge^, (iSize div 2) * iDir) ;
      Canvas.LineTo(X, Y) ;
      S := IntToStr(Abs(i DIV FDivisions)) ;
      Dec(PAxis^, TextSizeFunc(S) DIV 2) ;
      PEdge^ := ((iSize DIV 2) - ((iSize DIV 2) * (iEdge DIV iSize))) + iPadding ;
      Canvas.TextOut(X, Y, S);
    end
    else
    begin
      Inc(PEdge^, (iSize div 3) * iDir) ;
      Canvas.LineTo(X, Y) ;
    end ; { if }
    Inc(i) ;
  until(PAxis^ >= iMax) ;
  PositionShowing := False ;
end;

procedure TGRuler.SetAlignment(const Value: TAlignment);
begin
  FAlignment := Value;
  Invalidate ;
end;

procedure TGRuler.SetDivisions(const Value: Integer);
begin
  FDivisions := Value;
  Invalidate ;
end;

procedure TGRuler.SetEdge(const Value: TRulerEdge);
begin
  FEdge := Value;
  if(FOrientation = roHorizontal) then
    case FEdge of
      reRight : FEdge := reBottom ;
      reLeft : FEdge := reTop ;
    end { case }
  else
    case FEdge of
      reBottom : FEdge := reRight ;
      reTop : FEdge := reLeft ;
    end ; { case }
  Invalidate ;
end;

procedure TGRuler.SetMeasureUnit(const Value: TMeasureUnit);
begin
  FUnit := Value;
  case FUnit of
    muInches : FPixelsPerUnit := Screen.PixelsPerInch ;
    muCentimeters : FPixelsPerUnit := Trunc(Screen.PixelsPerInch / 2.54) ;
  end ; { case }
  Invalidate ;
end;

procedure TGRuler.SetOffset(const Value: Integer);
begin
  FOffset := Value;
  Invalidate ;
end;

procedure TGRuler.SetOrientation(const Value: TRulerOrientation);
var
  i : Integer ;
begin
  if((Value <> FOrientation) AND (NOT (csLoading in ComponentState))) then
  begin
    i := Width ;
    Width := Height ;
    Height := i ;
  end ; { if }
  FOrientation := Value;
  Edge := FEdge ;
  Invalidate ;
end;

procedure TGRuler.SetPixelsPerUnit(const Value: Integer);
begin
  FPixelsPerUnit := Value;
  if(Value = Screen.PixelsPerInch) then
    FUnit := muInches
  else
    if(Value = Trunc(Screen.PixelsPerInch / 2.54)) then
      FUnit := muCentimeters
    else
      FUnit := muCustom ;
  Invalidate ;
end;

procedure TGRuler.SetShowPosition(const Value: Boolean);
begin
  if(FShowPosition AND (NOT Value)) then
    Invalidate ;
  FShowPosition := Value;
end;

procedure TGRuler.SetTickColor(const Value: TColor);
begin
  FTickColor := Value ;
  Invalidate ;
end;

function TGRuler.StorePixelsPerUnit: Boolean;
begin
  Result := (FUnit = muCustom) ;
end;

procedure TGRuler.UpdatePosition(const X, Y: Integer);
var
  ptStart, ptEnd : TPoint ;
  procedure DrawLine ;
  begin
    Canvas.MoveTo(ptStart.X, ptStart.Y);
    Canvas.LineTo(ptEnd.X, ptEnd.Y) ;
  end ;
begin
  if(NOT FShowPosition) then
    exit ;
  Canvas.Pen.Color := clBlack ;
  Canvas.Pen.Mode := pmNotXor ;
  if(FOrientation = roHorizontal) then
  begin
    if(PositionShowing) then
    begin
      Canvas.MoveTo(LastPosition, 1) ;
      Canvas.LineTo(LastPosition, Height - 1);
    end ; { if }
    Canvas.MoveTo(X, 1) ;
    Canvas.LineTo(X, Height - 1) ;
    LastPosition := X ;
  end
  else
  begin
    if(PositionShowing) then
    begin
      Canvas.MoveTo(1, LastPosition) ;
      Canvas.LineTo(Width - 1, LastPosition);
    end ; { if }
    Canvas.MoveTo(1, Y) ;
    Canvas.LineTo(Width - 1, Y) ;
    LastPosition := Y ;
  end ; { if }
  Canvas.Pen.Mode := pmCopy ;
  PositionShowing := True ;
end;

end.
