unit QuickFindRecord;
{$R icon.dcr}
interface

uses
  Classes, ExtCtrls, Controls, Messages, StdCtrls, Graphics, ADODB_TLB;

const
  KEY_DELAY_TIME = 500;

  SCHEMA_COLUMN_NAME_FIELD = 'COLUMN_NAME';
  SCHEMA_DATA_TYPE_FIELD = 'DATA_TYPE';
  SCHEMA_TABLE_NAME_FIELD = 'TABLE_NAME';
  SCHEMA_LIMIT_REGULAR_TABLES = 'TABLE';

  DEFAULT_HEIGHT = 350;
  DEFAULT_WIDTH = 185;
  POINTS_PER_INCH = 72;
  EDIT_LIST_GAP = 6;

  LABELED_EDIT_IDE_NAME = 'EditBox';
  LIST_BOX_IDE_NAME = 'ListBox';
  DEFAULT_CAPTION = 'Type the first few letters:';

  TABLE_NOT_FOUND_ERR = 'Data table not found';
  FIELD_NOT_FOUND_ERR = 'Data Field not found';
  DISPLAY_FIELD_WRONG_TYPE = 'Search data field not string or string reference';
  INDEX_FIELD_ERR = 'Index field data type not 2 or 4 byte integer';
  CONNECTION_STRING_BLANK = 'Connection string not set';
  DISPLAY_TABLE_BLANK = 'Data table not set';
  DATA_CONNECTION_ERR = 'Error connecting to data or creating data objects';
  INVALID_INDEX_VALUE = -1;

type

   // For containing the acceptable/unacceptable data field types to check against user selected fields
  TDBTypeArray = array of integer;

  // sub class TListBox so we can hide multi-select prop (would fail if set to false)
  TListBoxEx = class(TListBox)
  private
    alwaysFalse: boolean;
  public
    constructor create(AOwner: TComponent); override;
  published
    property MultiSelect: boolean read alwaysFalse;
  end;

  //
  // This here component provides a way for a user to quickly navigate to the desired
  // record of a database table.
  // Works with extended characters, but I don't know how since I am just using regular string variables throughout
  //
  // It contains two sub components:
  // 1) A labeled edit.
  // 2) A list box.
  //
  // It also contains:
  // 1) An ADO connection
  // 2) An ADO recordset
  //
  // How it works:
  //    The first time it gets focus (and any other time it gets focus after either the TableName,
  //    DisplayField, or IndexField values change), it ensures that all those properties are groovy
  //    (are they really in the database, are the field types correct?)
  //    When the user types a key in the edit box, a thread is started that lasts for KEY_DELAY_TIME.
  //    When that time is up, if the user has typed another letter since, nothing happens.
  //    When the user pauses for KEY_DELAY_TIME, the text from the edit box is taken and queried
  //    against the DisplayField.
  //    If values are found, the list box is populated.  The first value whose first characters match
  //    what the user has typed is selected and the RecordSelected event fires.  If no mathches were
  //    found, the NothingSelected event fires (generally to clear out any controls that were populated
  //    when a previous RecordSelected event was fired).
  //    If the user arrows up or down, different values in the list box are selected, and the event fires
  //    again.
  //    After the event has fired, if the programmer has set the IndexField property, he may access the
  //    currentIndex property.  This will be the value of the IndexField that corresponds with the
  //    DisplayField value the user has selected.
  //
  // Usage:
  // 1) Set the value of TQuickFindFieldVal.DataQuery.ConnectionString if not set at design time.
  // 2) Set the TableName.
  // 3) Set the DisplayField to a string field in TableName that you want the values of to appear in the
  //    list below the edit box as the user types.
  // 4) Optionally set the value of IndexField to an integer field in TableName that is a key field you
  //    need to get other information from the database.
  // 5) Optionally set a TNotifyMethod to RecordSelected.
  // 6) Rock.
  //
  TQuickFindRecord = class(TCustomControl)
  private
    FEditBox: TLabeledEdit;
    mListBox: TListBoxEx;
    mCon: Connection;
    mRst: Recordset;
    mLabelHeight, mListHeight, mNanoSecondDelay: integer;
    FRecordSelected, FNothingSelected: TNotifyEvent;
    FTableName, FDisplayField, FIndexField, FConnectionString: string;
    mDisplayField, mIndexField: Field;
    mErr: shortstring;
    mValidIndexTypes, mInvalidDisplayTypes: TDBTypeArray;
    mIdentifiersValidated, mCtrlAltDown, mSearchFromStart, mJustEntered: boolean;
    mLastKeyUpTime: Cardinal;
    mLastKeyUp: Word;

    procedure onSize(var Message: TMessage); message WM_SIZE;
    procedure setTextAndSql;
    procedure setKeyResults(const currentText, sql: string);
    procedure moveListSelection(var Key: word);
    procedure setSelectedText;

    function  sqlSafe(const s: string): string;

    // First is run in a thread, second is called from thread
    procedure setSearchDelay;
    procedure initiateSearch;

//    procedure setField(const fieldName: string; theField: TField);
    function  getCurrentIndex: integer;
    procedure validateDBIdentifiers;

    // When any of these is set, set flag mIdentifiersValidated to false
    procedure setTableName(const Value: string);
    procedure setDisplayFieldName(const Value: string);
    procedure setIndexFieldName(const Value: string);

    procedure onEditKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure onEditKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure onListKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure onListClick(Sender: TObject);
  protected
    procedure Paint; override;
    procedure DoEnter; override;
  public
    constructor create(AOwner: TComponent); override;
    destructor  destroy; override;
    property  currentIndex: integer read getCurrentIndex;
    procedure manualSearch;
  published
    property EditBox: TLabeledEdit read FEditBox write FEditBox;
    property ListBox: TListBoxEx read mListBox write mListBox;
    property TableName: string read FTableName write setTableName;
    property DisplayField: string read FDisplayField write setDisplayFieldName;
    //
    // Optional:  If this field is set, then an index will be associated with every
    //  item in the result list (list box).  Then when the "RecordSelected" event
    //  fires, the user can read the value of "currentIndex" to get the index value
    //  associated with the selected displayField.
    //
    property IndexField: string read FIndexField write setIndexFieldName;
    //
    // When this is true, the filter will only return records with the text that
    // the user has typed where that text begins a field value (as opposed to the
    // text falling anywhere in the field value).  This will have the effect of
    // greatly limiting the match results (especially when only one or two characters are en-
    // tered), and speeding up the searches on keyup.
    //
    property FromStartOnly: boolean read mSearchFromStart write mSearchFromStart default false;
    property ConnectionString: string read FConnectionString write FConnectionString;
    property RecordSelected: TNotifyEvent read FRecordSelected write FRecordSelected;
    property NothingSelected: TNotifyEvent read FNothingSelected write FNothingSelected;
  end;

  procedure Register;

implementation

uses
  Windows, StrUtils, SysUtils, Math, Forms, JclStrings, Variants, CFXThreads, OleDB;

{ TQuickFindRecord }

procedure Register;
begin
  RegisterComponents('Colorflex', [TQuickFindRecord]);
end;

constructor TQuickFindRecord.create(AOwner: TComponent);
begin

  inherited Create(AOwner);

  Height := DEFAULT_HEIGHT;
  Width := DEFAULT_WIDTH;

  mSearchFromStart := false;
  mIdentifiersValidated := false;
  // convert our millisecond key delay to nanoseconds which is what the api timer wants
  // do this here so we don't calculate this on every key down
  mNanoSecondDelay := -(KEY_DELAY_TIME * 10000);

  // Add in all the database types to the set that we match against when validating the fields we use
  setLength(mValidIndexTypes, 4);
  // signed or unsigned 2 or 4 byte integer (can't be over 4 bytes because we're storing it' value in
  // the "object" property of a string list item as if it was a pointer to a memory address (which also
  // takes up 4 bytes).  We do this for speed
  mValidIndexTypes[0] := DBTYPE_I2;
  mValidIndexTypes[1] := DBTYPE_I4;
  mValidIndexTypes[2] := DBTYPE_UI2;
  mValidIndexTypes[3] := DBTYPE_UI4;
  setLength(mInvalidDisplayTypes, 4);
  mInvalidDisplayTypes[0] := DBTYPE_BSTR;
  mInvalidDisplayTypes[1] := DBTYPE_BYREF; // can be pointer to string
  mInvalidDisplayTypes[2] := DBTYPE_STR;
  mInvalidDisplayTypes[3] := DBTYPE_WSTR;

  // sub components
  FEditBox := TLabeledEdit.Create(self);
  FEditBox.Parent := self;
  FEditBox.Width := Width;
  FEditBox.OnKeyUp := onEditKeyUp;
  FEditBox.OnKeyDown := onEditKeyDown;

  mListBox := TListBoxEx.Create(self);
  mListBox.Parent := self;
  mListBox.Width := Width;
  mListBox.OnClick := onListClick;
  mListBox.OnKeyUp := onListKeyUp;

  mCon := CoConnection.Create;
  mCon.CursorLocation := adUseClient;
  mCon.Mode := adModeRead;
  mRst := CoRecordset.Create;

  if (csDesigning in ComponentState) then begin
    mListBox.SetSubComponent(true);
    FEditBox.SetSubComponent(true);
    FEditBox.Name := LABELED_EDIT_IDE_NAME;
    FEditBox.EditLabel.Caption := DEFAULT_CAPTION;
    mListBox.Name := LIST_BOX_IDE_NAME;
  end;
//  Color := clRed;
//  mListBox.Color := clblue;
//  FEditBox.EditLabel.Font.Size := 25;
//sendDebug('end create, address of feditbox: [' + strtohex('$' + inttostr(integer(@feditbox))) + ']');
end;

procedure TQuickFindRecord.moveListSelection(var Key: word);
//
// Determine which list item is already selected if any.
// Jump to the item below/above the currently selected one (top one if none already selected).
// Set the edit box's text to that item's.
// Select all in the edit box.
//
var
  selectedIndex: integer;
begin

  if mListBox.Items.Count > 0 then begin
    selectedIndex := mListBox.ItemIndex;

    case Key of
      // crazy nested ifThen:
      // First one checks whether an item was already selected.
      // Second one checks whether the selected one was at the end/beginning of the list
      VK_DOWN: selectedIndex := ifThen((selectedIndex = -1), 0, ifThen((selectedIndex = mListBox.Items.Count - 1), selectedIndex, selectedIndex + 1));
      VK_UP:   selectedIndex := ifThen((selectedIndex = -1), 0, ifThen((selectedIndex = 0), 0, selectedIndex - 1));
    end;

    mListBox.ItemIndex := selectedIndex;
    FEditBox.Text := mListBox.Items[selectedIndex];
    FEditBox.SelectAll;
    Key := 0;

    if assigned(FRecordSelected) then FRecordSelected(self);
  end;

end;

procedure TQuickFindRecord.onEditKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if ((Key = VK_DOWN) or (Key = VK_UP)) then
    moveListSelection(Key)
  //
  // We have to detect these keys on down because they might not be there on the key up
  // (when I do a control c, I roll my hand so the control is no longer there)
  //
  else if ((ssCtrl in Shift) or (ssAlt in Shift)) then
    mCtrlAltDown := true;
end;

procedure TQuickFindRecord.onEditKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin

  //
  // Ignore the first key up after the control has been entered.  This may be the key up of a tab
  // from tabbing in, or it may be the key up after the user has selected a menu item that creates
  // the form on which this component resides if this component's tab order is zero.
  //
  if mJustEntered then begin
    mJustEntered := false;
    exit;
  end;

  if (CharIsAlphaNum(chr(Key)) or (Key = VK_BACK) or (Key = VK_DELETE)) then begin
    if not mCtrlAltDown then begin
      mLastKeyUp := Key;
      mLastKeyUpTime := GetTickCount;  // set start time
      runThreadedMeth(setSearchDelay, initiateSearch);
    end else
      mCtrlAltDown := false;
  end;
  
end;

procedure TQuickFindRecord.onListClick(Sender: TObject);
begin
  setSelectedText;
end;

procedure TQuickFindRecord.onListKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if ((Key = VK_UP) or (Key = VK_DOWN)) then
    setSelectedText;
end;

procedure TQuickFindRecord.onSize(var Message: TMessage);
begin
  // Must set the edit box top below the label it contains because the TLabeledEdit
  // is not smart enough to do this (try dropping it on a form and positioning it to 0,0).
  // I can't set this on create because there is no canvas available then.
  // top of labeled edit = pixels per inch * (font size / 72)
  mLabelHeight := round(GetDeviceCaps(canvas.Handle, LOGPIXELSY) * (FEditBox.EditLabel.canvas.Font.Size / POINTS_PER_INCH)) + 3;
  FEditBox.Top := mLabelHeight;

  mListBox.Top := mLabelHeight + FEditBox.Height + EDIT_LIST_GAP;
  mListHeight := Height - mListBox.Top;
  inherited;
end;

procedure TQuickFindRecord.Paint;
begin
  // For some reason, setting the height in the onSize (which occurs before paint),
  // doesn't work there.

  // Somehow on one certain form, the TLabeledEdit subcomponent of this control got assigned to another
  // TLabeledEdit, and ours was nil during this event
//  if assigned(mListbox) then
//    sendDebug('mlistbox assigned')
//  else
//    sendDebug('mlistbox not assigned');
//  if assigned(feditbox) then
//    sendDebug('feditbox assigned')
//  else
//    sendDebug('feditbox not assigned');
  mListBox.Height := mListHeight;
  FEditBox.Width := Width;
  mListBox.Width := Width;
  inherited;
end;

//procedure TQuickFindRecord.setField(const fieldName: string; theField: TField);
//var
//  i: integer;
//  e: Exception;
//begin
//  if mQuery.Active then begin
//    for i := 0 to mQuery.FieldCount - 1 do begin
//      if mQuery.Fields[i].FieldName = fieldName then begin
//
//        //
//        // for index field, must be integer type field
//        // for display field, we don't care since we will display it's value calling the asString method anyway
//        //
//        if ((theField = FIndexField) and (not(mQuery.Fields[i] is TIntegerField))) then begin
//          e := Exception.Create(SET_INDEX_FIELD_ERR);
//          if (csDesigning in ComponentState) then
//            Application.ShowException(e)
//          else
//            raise e;
//        end;
//
//        theField := mQuery.Fields[i];
//        break;
//
//      end;
//    end;
//  end else
//    theField := nil;
//end;

procedure TQuickFindRecord.setKeyResults(const currentText, sql: string);
//
// I'm not going to separate any of this out into sub functions just to maximize
// speed.
// We're not going to do any exception handling here becuase it's slow and
// we validate the table/field names elsewhere.
//
var
  currentFieldVal: string;
  alreadySelected, completeSelectedText: boolean;
  i: integer;
begin

  mListBox.Clear;
  alreadySelected := false;
  i := 0;
  completeSelectedText := ((mLastKeyUp <> VK_DELETE) and (mLastKeyUp <> VK_BACK));

  //
  // Setting capacity of the list box requires getting the record count which requires
  // iterating through every record.  Not setting capacity makes adding the strings slower.
  // Time tests showed perhaps a slight advantage to not setting capacity.
  //
  mRst.Open(sql, mCon, adOpenForwardOnly, adLockReadOnly, adCmdText);
  if not (mRst.EOF and mRst.BOF) then begin
    mDisplayField := mRst.Fields.Item[FDisplayField];
    if FIndexField = '' then  // will raise error if the field doesn't exist
      mIndexField := nil
    else
      mIndexField := mRst.Fields.Item[FIndexField];

    while not mRst.Eof do begin

      currentFieldVal := string(mDisplayField.Value);
      mListBox.Items.Add(currentFieldVal);

      // If we're keeping track of indexes, stuff the value of the index in the current list item
      // (StringList)'s object field as if it was a 4 byte memory address.
      // We do our craziness for speed
      if assigned(mIndexField) then
        mListBox.Items.Objects[i] := Pointer(integer(mIndexField.Value));

      // Fill out the edit box with the full value of the first match leaving the part they haven't typed
      // yet selected for overwriting with consecutive keys.
      // Don't do this if they don't type an actual character (backspace would never work)
      if (not alreadySelected) and completeSelectedText then begin
        // if the current value's first letters match what the user has typed, fill out
        // the rest and select the part we filled out.
        if SameText(copy(currentFieldVal, 1, length(currentText)), currentText) then begin
          mListBox.Selected[i] := true;
          FEditBox.Text := currentFieldVal;
          FEditBox.SelStart := length(currentText);
          FEditBox.SelLength := length(currentFieldVal);
          alreadySelected := true;
        end;
      end;

      mRst.MoveNext;
      inc(i);
    end;
  end;

  // fire!
  if alreadySelected then begin
    if assigned(FRecordSelected) then FRecordSelected(self);
  end else
    if assigned(FNothingSelected) then FNothingSelected(self);

end;

procedure TQuickFindRecord.setSelectedText;
begin
  if mListBox.ItemIndex <> -1 then begin
    FEditBox.Text := mListBox.Items[mListBox.ItemIndex];
    FEditBox.SelectAll;
    FEditBox.SetFocus;
    if assigned(FRecordSelected) then FRecordSelected(self);
  end;
end;

function TQuickFindRecord.getCurrentIndex: integer;
begin
  if FIndexField <> '' then begin
    if (mListBox.ItemIndex = -1) then
      Result := INVALID_INDEX_VALUE
    else
      Result := integer(mListBox.Items.Objects[mListBox.ItemIndex]);
  end else
    Result := INVALID_INDEX_VALUE;
end;

procedure TQuickFindRecord.manualSearch;
//
// Call this if you're setting the text of the edit box yourself (as opposed to the user typing it
// in), and you want to start the search without having to simulate a key event (along with the delay
// that comes with it).
//
begin
  try
    validateDBIdentifiers;
    setTextAndSql;
  except
    ;
  end;
end;

function TQuickFindRecord.sqlSafe(const s: string): string;
// Duplicate all single quotes so sql search statement passes
var
  loc, position: integer;
begin
  result := s;
  loc := pos(#39, result);
  position := loc + 2;
  while (loc > 0) do begin
    Result := copy(result, 1, loc) + #39 + copy(result, loc + 1, length(result));
    loc := PosEx(#39, result, position);
    position := loc + 2;
  end;
end;

{ TListBoxEx }

constructor TListBoxEx.create(AOwner: TComponent);
begin
  inherited create(AOwner);
  alwaysFalse := false;
  inherited MultiSelect := alwaysFalse;
end;

destructor TQuickFindRecord.destroy;
begin
  if mRst.State = adStateOpen then mRst.Close;
  if mCon.State = adStateOpen then mCon.Close;
  mRst := nil;
  mCon := nil;
  inherited;
end;

procedure TQuickFindRecord.validateDBIdentifiers;
// when control is activated, we have to ensure:
// 1) Table name exists in database
// 2) Display field exists in table
// 3) Index field (if set) exists in table and is an integer field
// We don't want to do this on every key up event.
// If anything isn't valid, then disable the edit box.
var
  ole: OLEVariant;

  function invalidType(currentType: integer; const typeArray: TDBTypeArray): boolean;
  var
    i: integer;
  begin
    Result := true;
    for i := low(typeArray) to high(typeArray) do begin
      if currentType = typeArray[i] then begin
        Result := false;
        exit;
      end;
    end;
  end;

  function tableFound: boolean;
  var
    fld: Field;
  begin
    Result := false;

    ole := VarArrayCreate([0,3], varVariant);
    ole[3] := SCHEMA_LIMIT_REGULAR_TABLES;
    mRst := mCon.OpenSchema(adSchemaTables, ole, EmptyParam);
//    mRst.MoveFirst;
    fld := mRst.Fields.Item[SCHEMA_TABLE_NAME_FIELD];

    while not mRst.EOF do begin
      if SameText(string(fld.Value), FTableName) then begin
        Result := true;
        break;
      end;
      mRst.MoveNext;
    end;

    mRst.Close;
    if not result then
      mErr := TABLE_NOT_FOUND_ERR;
  end;

  function fieldGood(const fieldName: string; isInt: boolean): boolean;
  //
  // Validate each field name:
  // Ensure it's actually in the selected table and that it's of an acceptable type.
  //
  var
    currentType: integer;
    nameField, typeField: Field;
  begin
    ole := VarArrayCreate([0,2], varVariant);
    ole[2] := FTableName;
    mRst := mCon.OpenSchema(adSchemaColumns, ole, EmptyParam);
    with mRst do begin
      nameField := Fields.Item[SCHEMA_COLUMN_NAME_FIELD];
      typeField := Fields.Item[SCHEMA_DATA_TYPE_FIELD];
//      MoveFirst;
      while not eof do begin

        // data engines don't seem to care about case in table names
        if SameText(string(nameField.Value), fieldName) then begin
          currentType := integer(typeField.Value);
          if isInt then begin
            if invalidType(currentType, mValidIndexTypes) then
              mErr := INDEX_FIELD_ERR;
          end else begin
            if invalidType(currentType, mInvalidDisplayTypes) then
              mErr := DISPLAY_FIELD_WRONG_TYPE;
          end;
          break;
        end;

        MoveNext;
      end;

      // if we're at eof we never found the main display field
      if eof then mErr := FIELD_NOT_FOUND_ERR;
    end;
    Result := mErr = '';
  end;

begin

  // init
  if assigned(mRst) then begin
    if mRst.State = adStateOpen then mRst.Close;
  end;
  if mCon.State = adStateOpen then mCon.Close;
  mErr := '';

  // checks...
  if FConnectionString <> '' then begin
    if FDisplayField <> '' then begin
      try
        mCon.Open(FConnectionString, '', '', integer(adConnectUnspecified));
        if tableFound then begin
          if fieldGood(FDisplayField, false) then begin
            if FIndexField = '' then
              mIdentifiersValidated := true
            else
              mIdentifiersValidated := fieldGood(FIndexField, true);
            // Create a new recordset (props of one returned by open schema not what we want)
            // ref counting should leave no leaks
            mRst := CoRecordset.Create;
          end;
        end;
      except
        mErr := DATA_CONNECTION_ERR;
      end;
    end else
      mErr := DISPLAY_TABLE_BLANK;
  end else
    mErr := CONNECTION_STRING_BLANK;

  // disable if necessary
  if not mIdentifiersValidated then begin
    mListBox.Clear;
    mListBox.Items.Add(mErr);
    mListBox.Enabled := false;
    FEditBox.Enabled := false;
  end;

end;

procedure TQuickFindRecord.DoEnter;
//
// Validate the selected table/fields when the control is activated if not already done.
// Setting of table name, display field name, or index field name will set flag validateDBIdentifiers false
//
begin
  inherited;
  mJustEntered := true;
  if not mIdentifiersValidated then
    validateDBIdentifiers;
end;

procedure TQuickFindRecord.setDisplayFieldName(const Value: string);
begin
  FDisplayField := Value;
  mIdentifiersValidated := false;
end;

procedure TQuickFindRecord.setIndexFieldName(const Value: string);
begin
  FIndexField := Value;
  mIdentifiersValidated := false;
end;

procedure TQuickFindRecord.setTableName(const Value: string);
begin
  FTableName := Value;
  mIdentifiersValidated := false;
end;

procedure TQuickFindRecord.setTextAndSql;
var
  currentText, currentSql: string;
begin

  currentText := sqlSafe(FEditBox.Text);
  if mRst.State = adStateOpen then mRst.Close;

  // set the first sql string (a string should always be at index 0 because we added one on create
  // setting faster than clearing and adding
  if FIndexField = '' then
    currentSql := 'Select ' + FDisplayField + ' From ' + FTableName
  else
    currentSql := 'Select ' + FIndexField + ',' + FDisplayField + ' From ' + FTableName;
  if mSearchFromStart then
    currentSql := currentSql + ' Where ' + FDisplayField + ' LIKE ''' + currentText + '%'''
  else
    currentSql := currentSql + ' Where ' + FDisplayField + ' LIKE ''%' + currentText + '%''';

  mIndexField := nil;
  mDisplayField := nil;
  setKeyResults(FEditBox.Text, currentSql);

end;

procedure TQuickFindRecord.initiateSearch;
//
// setSearchTimer is run in a separate thread.  When it finishes, it calls this method
// which checks to see if another key has been pressed within the last KEY_DELAY_TIME.
// If so, nothing happens; if not, we begin the search
//
begin
  if ((GetTickCount - mLastKeyUpTime) >= KEY_DELAY_TIME) then
    setTextAndSql;
end;

procedure TQuickFindRecord.setSearchDelay;
var
  duhTime: TLargeInteger;
  timerHandle: THandle;
begin
  timerHandle := CreateWaitableTimer(nil, false, '');
  if ((timerHandle = INVALID_HANDLE_VALUE) or (timerHandle = ERROR_ALREADY_EXISTS)) then
    exit;

  Large_Integer(duhTime).QuadPart := mNanoSecondDelay;

  if SetWaitableTimer(timerHandle, duhTime, 0, nil, nil, LongBool(false)) then
    WaitForSingleObject(timerHandle, INFINITE);

  CloseHandle(timerHandle);
end;

end.
